unit uComEvents;
{
  COM Event enabling
  Copyright  1997 Run-Time Systems
  Public and commerical use is granted provided this header is not removed.

  Some minor modifications were made to this file by Brian Dupras (not affiliated
  with Run-Time Systems in any way).
}
interface
//
// TObjectObjectWithEvents inherited classes need to:
//    1. Define a event dispinterface (ie. IObjectEvents) in the .tlb
//    2. Change TAutoObject to TAutoObjectWithEvents in generated impl
//    3. Change TAutoObjectFactory to TAutoObjectWithEventsFactory
//    4. Add an IObjectsEvents field to the object declaration
//    5. Override abstract EventSinkChanged as follows:
//          procedure TEventObject.EventSinkChanged( const Events: IUnknown );
//          begin
//            FEvents := Events as IObjectEvents;
//          end;
//    6. Add public <<EventName>>Event procedures to fire the event
//          ie. procedure OnCloseEvent;
//              begin
//                if Assigned(FEvents) then FEvents.OnClose;
//              end;
//    7. Call the events whenever appropriate
//

uses  Windows
    , ComObj
    , ActiveX
    , AxCtrls     // IConnectionPointContainer (can roll your own)
    , uComExtra   // interface files
    //commented by Brian Dupras //, uComExtra2  // implementation files
    ;

type
  TAutoObjectWithEventsFactory = class;
  TAutoObjectWithEvents = class(TAutoObject, IObjectSafety)
  protected // should be private
    FEventFactory: TAutoObjectWithEventsFactory;
    FEventSink: IUnknown;
    FConnectionPoints: TConnectionPoints;
  protected
    FUseDispatchSilencer: Boolean;
  protected
    procedure EventSinkChanged( const Events: IUnknown ); virtual; abstract;
    procedure EventConnect( const Sink: IUnknown; Connecting: Boolean ); virtual;
    function  ObjQueryInterface( const iid: TGUID; out Obj ): HResult; override;
    function  GetInterfaceSafetyOptions( const iid: TGUID;
                                         out dwSupportedOptions: DWord;
                                         out dwEnabledOptions: DWord ): HResult; stdcall;
    function  SetInterfaceSafetyOptions( const iid: TGUID;
                                         dwOptionSetMask: DWord;
                                         dwEnabledOptions: DWord ): HResult; stdcall;
  public
    procedure Initialize; override;
    destructor  Destroy; override;
  public
    property  EventFactory: TAutoObjectWithEventsFactory read FEventFactory;
  end;

  TAutoObjectWithEventsClass = class of TAutoObjectWithEvents;


  // event enabling class factory
  TAutoObjectWithEventsFactory = class( TAutoObjectFactory )
  private
    FEventTypeInfo: ITypeInfo;
    FEventIID: TGUID;
  public
    constructor Create( ComServer: TComServerObject;
                        AutoClass: TAutoObjectWithEventsClass;
                        const ClassID: TGUID );
    destructor  Destroy; override;
  public
    property  EventIID: TGUID read FEventIID;
    property  EventTypeInfo: ITypeInfo read FEventTypeInfo;
  end;

{ TDispatchSilencer :: copied & exposed from AxCtrls }
type
  TDispatchSilencer = class(TInterfacedObject, IUnknown, IDispatch)
  private
    Dispatch: IDispatch;
    DispIntfIID: TGUID;
  public
    constructor Create(ADispatch: IUnknown; const ADispIntfIID: TGUID);
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): Integer; stdcall;
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): Integer; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): Integer; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): Integer; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): Integer; stdcall;
  end;

{ TUnknownSilencer }
type
  TUnknownSilencer = class( TInterfacedObject, IUnknown )
  private
    FIntf: IUnknown;
    FIID: TGUID;
  public
    constructor Create( Intf: IUnknown; const iid: TGUID);
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): Integer; stdcall;
  end;

implementation

uses SysUtils;

{ TAutoObjectWithEvents }
procedure TAutoObjectWithEvents.EventConnect( const Sink: IUnknown; Connecting: Boolean );
begin
  // answer a request for sinking
  if Connecting then begin
    OleCheck(Sink.QueryInterface( FEventFactory.FEventIID, FEventSink ));
    if FUseDispatchSilencer
      then EventSinkChanged( TDispatchSilencer.Create( Sink, FEventFactory.FEventIID ) )
      else EventSinkChanged( Sink );
  end else begin
    pointer(FEventSink) := nil;
    EventSinkChanged( nil );
  end;
end;

function  TAutoObjectWithEvents.ObjQueryInterface( const iid: TGUID; out Obj ): HResult;
begin
  Result := inherited ObjQueryInterface( iid, Obj );
  if Result <> 0 then begin
    if IsEqualIID( iid, IConnectionPointContainer ) then begin
      FConnectionPoints.GetInterface( iid, Obj );
      Result := S_OK;
    end; //if
  end; //if
end;

procedure TAutoObjectWithEvents.Initialize;
begin
  inherited Initialize;
  FEventFactory := TAutoObjectWithEventsFactory(Factory);
  FConnectionPoints := TConnectionPoints.Create( Self );

  // setup for events if we have events to hook to
  if (FEventFactory.FEventTypeInfo <> nil)
    then FConnectionPoints.CreateConnectionPoint(FEventFactory.FEventIID, ckSingle, EventConnect);
end;

destructor TAutoObjectWithEvents.Destroy;
begin
  // get rid of our connection point container (watch for _Release problems)
  FConnectionPoints.Free;
  inherited Destroy;
end;


function  TAutoObjectWithEvents.GetInterfaceSafetyOptions;
begin
  dwSupportedOptions := INTERFACESAFE_FOR_UNTRUSTED_CALLER or INTERFACESAFE_FOR_UNTRUSTED_DATA;
  dwEnabledOptions := dwSupportedOptions;
  Result := s_ok;
end;

function  TAutoObjectWithEvents.SetInterfaceSafetyOptions;
begin
  Result := s_ok;
end;

{ TDispatchSilencer }
constructor TDispatchSilencer.Create(ADispatch: IUnknown; const ADispIntfIID:TGUID);
begin
  inherited Create;
  DispIntfIID := ADispIntfIID;
  OleCheck(ADispatch.QueryInterface(ADispIntfIID, Dispatch));
end;

function TDispatchSilencer.QueryInterface(const IID:TGUID; out Obj):Integer;
begin
  Result:=inherited QueryInterface(IID, Obj);
  if (Result=E_NOINTERFACE)
    then if IsEqualGUID(IID, DispIntfIID) then begin
      IDispatch(Obj) := Self;
      Result := S_OK;
    end else Result := Dispatch.QueryInterface(IID, Obj);
end;

function TDispatchSilencer.GetTypeInfoCount(out Count: Integer): Integer;
begin
  Result:=Dispatch.GetTypeInfoCount(Count);
end;

function TDispatchSilencer.GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): Integer;
begin
  Result:=Dispatch.GetTypeInfo(Index, LocaleID, TypeInfo);
end;

function TDispatchSilencer.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): Integer;
begin
  Result:=Dispatch.GetIDsOfNames(IID, Names, NameCount, LocaleID, DispIDs);
end;

function TDispatchSilencer.Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
  Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): Integer;
begin
  { Ignore error since some containers, such as Internet Explorer 3.0x, will
    return error when the method was not handled, or scripting errors occur }
  Dispatch.Invoke(DispID, IID, LocaleID, Flags, Params, VarResult, ExcepInfo, ArgErr);
  Result := S_OK;
end;

{ TUnknownSilencer }
constructor TUnknownSilencer.Create( Intf: IUnknown; const iid: TGUID );
begin
  inherited Create;
  FIntf := Intf;
  FIID := iid;
end;
function TUnknownSilencer.QueryInterface(const IID: TGUID; out Obj): Integer;
begin
  Result := inherited QueryInterface( IID, Obj );
  if Result = E_NOINTERFACE then
    if IsEqualGUID( IID, FIID ) then begin
      IUnknown(Obj) := Self;
      Result := S_OK;
      end
    else
      Result := FIntf.QueryInterface( IID, Obj );
end;


{ TAutoObjectWithEventsFactory }
constructor TAutoObjectWithEventsFactory.Create(
                ComServer: TComServerObject;
                AutoClass: TAutoObjectWithEventsClass;
                const ClassID: TGUID );
var
  TypeAttr: PTypeAttr;
begin
  inherited Create( ComServer, AutoClass, ClassID, ciMultiInstance );
  FEventTypeInfo := GetInterfaceTypeInfo(
                        ImplTypeFlag_FDefault or ImplTypeFlag_FSource );
  if Assigned(FEventTypeInfo) then begin
    OleCheck(FEventTypeInfo.GetTypeAttr( TypeAttr ));
    FEventIID := TypeAttr.guid;
    FEventTypeInfo.ReleaseTypeAttr( TypeAttr );
    end;
end;
destructor  TAutoObjectWithEventsFactory.Destroy;
begin
  inherited Destroy;
end;

end.

